""" Constant modules

Here are defined the most common constants used in DarkNews.

We use PDG2020 values for constants and SM masses.

Some low-level auxiliary functions are defined.

"""
import subprocess
import numpy as np
from numpy import sqrt
import math
from scipy import interpolate

try:
    import importlib.resources as resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as resources

################################################
# constants of light cm/s
c_LIGHT = 29_979_245_800
hb = 6.582119569e-25  # hbar in Gev s

m_proton_in_kg = 1.6726219236951e-27
m_proton_in_g = m_proton_in_kg * 1e3
m_proton_in_t = m_proton_in_kg * 1e-3

################################################
# constants for convertion
MeV_to_GeV = 1e-3

invm2_to_incm2 = 1e-4
fb_to_cm2 = 1e-39
NAvo = 6.02214076 * 1e23
rad_to_deg = 180.0 / np.pi
deg_to_rad = 1 / rad_to_deg

invGeV_to_cm = hb * c_LIGHT  # hbar c = 197.3269804e-16 GeV.cm
invGeV2_to_cm2 = invGeV_to_cm ** 2

fm_to_GeV = 1 / invGeV_to_cm * 1e-15 * 1e2
invGeV_to_s = invGeV_to_cm / c_LIGHT

cm2_to_attobarn = 1e42
attobarn_to_cm2 = 1e-42

invGeV2_to_attobarn = invGeV2_to_cm2 * cm2_to_attobarn

GeV2_to_cm3s = invGeV2_to_cm2 * c_LIGHT
invcm3_to_eV3 = invGeV2_to_cm2 ** (3 / 2) * 1e27

g_to_eV = 5.6095886031e32  # eV
kg_to_eV = 5.6095886031e35  # eV
t_to_eV = 5.6095886031e38  # eV

g_to_GeV = g_to_eV * 1e-9  # GeV
kg_to_GeV = kg_to_eV * 1e-9  # GeV
t_to_GeV = t_to_eV * 1e-9  # GeV

#####################################
# indices of neutral leptons
ind_e = 0
ind_mu = 1
ind_tau = 2

# list of active neutrino indices
inds_active = range(ind_tau + 1)

N4 = "N4"
N5 = "N5"
N6 = "N6"

################################################
# Masses

# quarks
m_u = 2.16e-3  # GeV
m_d = 4.67e-3  # GeV
m_s = 93e-3  # GeV
m_c = 1.27  # GeV
m_b = 4.18  # GeV
m_t = 172.76  # GeV

# nucleons
m_proton = 0.93827208816  # GeV
m_neutron = 0.93956542052  # GeV
m_avg = (m_proton + m_neutron) / 2.0  # GeV

# leptons
m_W = 80.37912  # GeV
m_Z = 91.187621  # GeV

m_e = 0.5109989500015e-3  # GeV
m_mu = 0.1056583755  # GeV
m_tau = 1.77686  # GeV

# charged hadrons
m_charged_pion = 0.1396
m_charged_rho = 0.7758

# neutral hadrons
m_neutral_pion = 0.135
m_neutral_eta = 0.5478
m_neutral_rho = 0.7755

m_neutral_B = 5.27958
m_charged_B = 5.27958

m_neutral_kaon = 0.497611
m_charged_kaon = 0.4937
m_charged_kaonstar = 0.892

################################################
# QED
alphaQED = 1.0 / 137.03599908421  # Fine structure constant at q2 -> 0
eQED = np.sqrt((4 * np.pi) * alphaQED)

# get running alphaQED
Q, inv_alphaQED = np.genfromtxt(resources.open_text("DarkNews.include.aux_data","alpha_QED_running_posQ2.dat"),
                                 unpack=True)
runningAlphaQED = interpolate.interp1d(Q, 1.0 / inv_alphaQED)

################################################
# Weak sector
Gf = 1.16637876e-5  # Fermi constant (GeV^-2)
gweak = np.sqrt(Gf * m_W ** 2 * 8 / np.sqrt(2))
s2w = 0.22343  # On-shell
sw = np.sqrt(s2w)
cw = np.sqrt(1.0 - s2w)

################################################
# Higgs -- https://pdg.lbl.gov/2019/reviews/rpp2018-rev-higgs-boson.pdf
vev_EW = 1 / np.sqrt(np.sqrt(2) * Gf)
m_H = 125.10
lamb_quartic = (m_H / vev_EW) ** 2 / 2
m_h_potential = -lamb_quartic * vev_EW ** 2

################################################
# Mesons
fcharged_pion = 0.1307
fcharged_kaon = 0.1598
fcharged_rho = 0.220

fneutral_pion = 0.130
fneutral_kaon = 0.164
fneutral_B = 0.1909
fcharged_B = 0.1871
fneutral_Bs = 0.2272
fneutral_eta = 0.210

Fneutral_pion = fneutral_pion / np.sqrt(2.0)
Fneutral_kaon = fneutral_kaon / np.sqrt(2.0)
Fneutral_B = fneutral_B / np.sqrt(2.0)
Fcharged_B = fcharged_B / np.sqrt(2.0)
Fneutral_Bs = fneutral_Bs / np.sqrt(2.0)
Fneutral_eta = fneutral_eta / np.sqrt(2.0)


################################################
# CKM elements
# PDG2019
lamCKM = 0.22453
ACKM = 0.836
rhoBARCKM = 0.122
etaBARCKM = 0.355
rhoCKM = rhoBARCKM / (1 - lamCKM * lamCKM / 2.0)
etaCKM = etaBARCKM / (1 - lamCKM * lamCKM / 2.0)

s12 = lamCKM
s23 = ACKM * lamCKM ** 2
s13e = (
    ACKM
    * lamCKM ** 3
    * (rhoCKM + 1j * etaCKM)
    * np.sqrt(1.0 - ACKM ** 2 * lamCKM ** 4)
    / (np.sqrt(1.0 - lamCKM ** 2) * (1.0 - ACKM ** 2 * lamCKM ** 4 * (rhoCKM + 1j * etaCKM)))
)
c12 = np.sqrt(1 - s12 ** 2)
c23 = np.sqrt(1 - s23 ** 2)
c13 = np.sqrt(1 - abs(s13e) ** 2)
Vud = c12 * c13
Vus = s12 * c13
Vub = np.conj(s13e)
Vcd = -s12 * c23 - c12 * s23 * s13e
Vcs = c12 * c23 - s12 * s23 * s13e
Vcb = s23 * c13
Vtd = s12 * s23 - c12 * c23 * s13e
Vts = -c12 * s23 - s12 * c23 * s13e
Vtb = c23 * c13


################################################
# PMNS parameters
# NuFit Oct 2021 -- http://www.nu-fit.org/?q=node/238#label85

s12 = np.sqrt(0.304)
s23 = np.sqrt(0.450)
s13 = np.sqrt(0.02246)
c12 = np.sqrt(1 - s12 ** 2)
c23 = np.sqrt(1 - s23 ** 2)
c13 = np.sqrt(1 - s13 ** 2)
delta = 230 * deg_to_rad

Ue1 = c12 * c13
Ue2 = s12 * c13
Ue3 = s13 * np.exp(-delta * 1j)
Umu1 = -s12 * c23 - c12 * s23 * s13 * np.exp(delta * 1j)
Umu2 = c12 * c23 - s12 * s23 * s13 * np.exp(delta * 1j)
Umu3 = s23 * c13
Utau1 = s12 * s23 - c12 * c23 * s13 * np.exp(delta * 1j)
Utau2 = -c12 * s23 - s12 * c23 * s13 * np.exp(delta * 1j)
Utau3 = c23 * c13
UPMNS = np.matrix([[Ue1, Ue2, Ue3], [Umu1, Umu2, Umu3], [Utau1, Utau2, Utau3]], dtype=complex)


################################################
# low-level auxiliary functions
################################################
def is_odd(num):
    return num & 0x1


np_is_odd = np.vectorize(is_odd)

# https://newbedev.com/how-do-i-annotate-with-power-of-ten-formatting vvv
# Define function for string formatting of scientific notation
def sci_notation(num, decimal_digits=1, precision=None, exponent=None):
    """
    Returns a string representation of the scientific
    notation of the given number formatted for use with
    LaTeX or Mathtext, with specified number of significant
    decimal digits and precision (number of decimal digits
    to show). The exponent to be used can also be specified
    explicitly.
    """
    if num != 0:
        if exponent is None:
            exponent = int(math.floor(math.log10(abs(num))))
        coeff = round(num / float(10 ** exponent), decimal_digits)
        if precision is None:
            precision = decimal_digits

        return r"${0:.{2}f}\cdot10^{{{1:d}}}$".format(coeff, exponent, precision)
    else:
        return r"0"


# ^^^

# run shell commands from notebook
def subprocess_cmd(command, verbose=2):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = process.communicate()
    if verbose == 2:
        print(command)
        print(stdout.decode("utf-8"))
        print(stderr.decode("utf-8"))
    elif verbose == 1:
        if len(stderr.decode("utf-8")) > 2:
            print(command)
            print("n", stderr.decode("utf-8"), "m")


def get_decay_rate_in_s(G):
    return 1.0 / G * invGeV_to_s


def get_decay_rate_in_cm(G):
    return 1.0 / G * invGeV_to_cm


# phase space function
def kallen(a, b, c):
    return (a - b - c) ** 2 - 4 * b * c


def kallen_sqrt(a, b, c):
    return np.sqrt(kallen(a, b, c))


#### aliases used in MATHEMATICA copy-paste
MZBOSON = m_Z
MW = m_W

# from CForm to pythonic syntax
Power = lambda x, n: x ** n
Sqrt = lambda x: sqrt(x)
Pi = np.pi


import argparse


def add_common_bsm_arguments(parser, DEFAULTS):

    #### name of the generation and model
    parser.add_argument("--name", type=str, help="name of the generation and model")

    ##### file containing the parameters
    parser.add_argument("--param-file", type=str, help="file containing parameters definitions")

    ##### dark sector spectrum
    parser.add_argument("--mzprime", type=float, help="Z' mass")
    parser.add_argument("--mhprime", type=float, help="h' mass")
    parser.add_argument("--m4", type=float, help="mass of the fourth neutrino")
    parser.add_argument("--m5", type=float, help="mass of the fifth neutrino")
    parser.add_argument("--m6", type=float, help="mass of the sixth neutrino")

    parser.add_argument("--HNLtype", type=str.lower, help="HNLtype: dirac or majorana", choices=DEFAULTS._choices["HNLtype"])

    ##### TMM in GeV^-1
    parser.add_argument("--mu_tr_e4", type=float, help="TMM mu_tr_e4")
    parser.add_argument("--mu_tr_e5", type=float, help="TMM mu_tr_e5")
    parser.add_argument("--mu_tr_e6", type=float, help="TMM mu_tr_e6")

    parser.add_argument("--mu_tr_mu4", type=float, help="TMM mu_tr_mu4")
    parser.add_argument("--mu_tr_mu5", type=float, help="TMM mu_tr_mu5")
    parser.add_argument("--mu_tr_mu6", type=float, help="TMM mu_tr_mu6")

    parser.add_argument("--mu_tr_tau4", type=float, help="TMM mu_tr_tau4")
    parser.add_argument("--mu_tr_tau5", type=float, help="TMM mu_tr_tau5")
    parser.add_argument("--mu_tr_tau6", type=float, help="TMM mu_tr_tau6")

    parser.add_argument("--mu_tr_44", type=float, help="TMM mu_tr_44")
    parser.add_argument("--mu_tr_45", type=float, help="TMM mu_tr_45")
    parser.add_argument("--mu_tr_46", type=float, help="TMM mu_tr_46")

    parser.add_argument("--mu_tr_55", type=float, help="TMM mu_tr_55")
    parser.add_argument("--mu_tr_56", type=float, help="TMM mu_tr_56")

    parser.add_argument("--mu_tr_66", type=float, help="TMM mu_tr_66")

    ##### Scalar vertices
    parser.add_argument("--s_e4", type=float, help="scalar vertex s_e4")
    parser.add_argument("--s_e5", type=float, help="scalar vertex s_e5")
    parser.add_argument("--s_e6", type=float, help="scalar vertex s_e6")

    parser.add_argument("--s_mu4", type=float, help="scalar vertex s_mu4")
    parser.add_argument("--s_mu5", type=float, help="scalar vertex s_mu5")
    parser.add_argument("--s_mu6", type=float, help="scalar vertex s_mu6")

    parser.add_argument("--s_tau4", type=float, help="scalar vertex s_tau4")
    parser.add_argument("--s_tau5", type=float, help="scalar vertex s_tau5")
    parser.add_argument("--s_tau6", type=float, help="scalar vertex s_tau6")

    parser.add_argument("--s_44", type=float, help="scalar vertex s_44")
    parser.add_argument("--s_45", type=float, help="scalar vertex s_45")
    parser.add_argument("--s_46", type=float, help="scalar vertex s_46")

    parser.add_argument("--s_55", type=float, help="scalar vertex s_55")
    parser.add_argument("--s_56", type=float, help="scalar vertex s_56")

    parser.add_argument("--s_66", type=float, help="scalar vertex s_66")


def add_three_portal_arguments(parser, DEFAULTS):
    ##### neutral lepton mixing
    parser.add_argument("--Ue4", type=float, help="Ue4")
    parser.add_argument("--Ue5", type=float, help="Ue5")
    parser.add_argument("--Ue6", type=float, help="Ue6")

    parser.add_argument("--Umu4", type=float, help="Umu4")
    parser.add_argument("--Umu5", type=float, help="Umu5")
    parser.add_argument("--Umu6", type=float, help="Umu6")

    parser.add_argument("--Utau4", type=float, help="Utau4")
    parser.add_argument("--Utau5", type=float, help="Utau5")
    parser.add_argument("--Utau6", type=float, help="Utau6")

    parser.add_argument("--UD4", type=float, help="UD4")
    parser.add_argument("--UD5", type=float, help="UD5")
    parser.add_argument("--UD6", type=float, help="UD6")

    ##### dark coupling choices
    parser.add_argument("--gD", type=float, help="U(1)_d dark coupling")
    parser.add_argument("--alphaD", type=float, help="U(1)_d  alpha_dark = (g_dark^2 /4 pi)")

    ##### kinetic mixing options
    parser.add_argument("--epsilon", type=float, help="epsilon")
    parser.add_argument("--epsilon2", type=float, help="epsilon^2")
    parser.add_argument("--alpha_epsilon2", type=float, help="alpha_QED*epsilon^2")
    parser.add_argument("--chi", type=float, help="chi")

    parser.add_argument("--theta", type=float, help="Scalar mixing angle between h-h'")


def add_generic_bsm_arguments(parser, DEFAULTS):

    ##### Z boson couplings
    parser.add_argument("--c_e4", type=float, help="SM Z boson vertex c_e4 ")
    parser.add_argument("--c_e5", type=float, help="SM Z boson vertex c_e5 ")
    parser.add_argument("--c_e6", type=float, help="SM Z boson vertex c_e6 ")
    parser.add_argument("--c_mu4", type=float, help="SM Z boson vertex c_mu4 ")
    parser.add_argument("--c_mu5", type=float, help="SM Z boson vertex c_mu5 ")
    parser.add_argument("--c_mu6", type=float, help="SM Z boson vertex c_mu6 ")
    parser.add_argument("--c_tau4", type=float, help="SM Z boson vertex c_tau4 ")
    parser.add_argument("--c_tau5", type=float, help="SM Z boson vertex c_tau5 ")
    parser.add_argument("--c_tau6", type=float, help="SM Z boson vertex c_tau6 ")
    parser.add_argument("--c_44", type=float, help="SM Z boson vertex c_44 ")
    parser.add_argument("--c_45", type=float, help="SM Z boson vertex c_45 ")
    parser.add_argument("--c_46", type=float, help="SM Z boson vertex c_46 ")
    parser.add_argument("--c_55", type=float, help="SM Z boson vertex c_55 ")
    parser.add_argument("--c_56", type=float, help="SM Z boson vertex c_56 ")
    parser.add_argument("--c_66", type=float, help="SM Z boson vertex c_66 ")

    ##### Z' vector couplings
    parser.add_argument("--d_e4", type=float, help="Z' vector vertex d_e4 ")
    parser.add_argument("--d_e5", type=float, help="Z' vector vertex d_e5 ")
    parser.add_argument("--d_e6", type=float, help="Z' vector vertex d_e6 ")
    parser.add_argument("--d_mu4", type=float, help="Z' vector vertex d_mu4 ")
    parser.add_argument("--d_mu5", type=float, help="Z' vector vertex d_mu5 ")
    parser.add_argument("--d_mu6", type=float, help="Z' vector vertex d_mu6 ")
    parser.add_argument("--d_tau4", type=float, help="Z' vector vertex d_tau4 ")
    parser.add_argument("--d_tau5", type=float, help="Z' vector vertex d_tau5 ")
    parser.add_argument("--d_tau6", type=float, help="Z' vector vertex d_tau6 ")
    parser.add_argument("--d_44", type=float, help="Z' vector vertex d_44 ")
    parser.add_argument("--d_45", type=float, help="Z' vector vertex d_45 ")
    parser.add_argument("--d_46", type=float, help="Z' vector vertex d_46 ")
    parser.add_argument("--d_55", type=float, help="Z' vector vertex d_55 ")
    parser.add_argument("--d_56", type=float, help="Z' vector vertex d_56 ")
    parser.add_argument("--d_66", type=float, help="Z' vector vertex d_66 ")

    ##### Charge particle couplings
    parser.add_argument("--ceV", type=float, help="SM Z boson vector vertex to charged leptons ceV")
    parser.add_argument("--ceA", type=float, help="SM Z boson axial vertex to charged leptons ceA")
    parser.add_argument("--cuV", type=float, help="SM Z boson vector vertex to up quark cuV")
    parser.add_argument("--cuA", type=float, help="SM Z boson axial vertex to up quark cuA")
    parser.add_argument("--cdV", type=float, help="SM Z boson vector vertex to down quark cdV")
    parser.add_argument("--cdA", type=float, help="SM Z boson axial vertex to down quark cdA")

    parser.add_argument("--deV", type=float, help="Z' vector vertex to charged leptons deV")
    parser.add_argument("--deA", type=float, help="Z' axial vertex to charged leptons deA")
    parser.add_argument("--duV", type=float, help="Z' vector vertex to up quark duV")
    parser.add_argument("--duA", type=float, help="Z' axial vertex to up quark duA")
    parser.add_argument("--ddV", type=float, help="Z' vector vertex to down quark ddV")
    parser.add_argument("--ddA", type=float, help="Z' axial vertex to down quark ddA")

    parser.add_argument("--deS", type=float, help="h' scalar vertex to charged leptons deS")
    parser.add_argument("--deP", type=float, help="h' pseudoscalar vertex to charged leptons deP")
    parser.add_argument("--duS", type=float, help="h' scalar vertex to up quark duS")
    parser.add_argument("--duP", type=float, help="h' pseudoscalar vertex to up quark duP")
    parser.add_argument("--ddS", type=float, help="h' scalar vertex to down quark ddS")
    parser.add_argument("--ddP", type=float, help="h' pseudoscalar vertex to down quark ddP")

    parser.add_argument("--cprotonV", type=float, help="SM Z boson vector vertex to protons cpV")
    parser.add_argument("--cneutronV", type=float, help="SM Z boson axial vertex to neutrons cnV")

    parser.add_argument("--cprotonA", type=float, help="SM Z boson vector vertex to protons cpA")
    parser.add_argument("--cneutronA", type=float, help="SM Z boson axial vertex to neutrons cnA")

    parser.add_argument("--dprotonV", type=float, help="SM Z boson vector vertex to protons dpV")
    parser.add_argument("--dneutronV", type=float, help="SM Z boson axial vertex to neutrons dnV")

    parser.add_argument("--dprotonA", type=float, help="SM Z boson vector vertex to protons dpA")
    parser.add_argument("--dneutronA", type=float, help="SM Z boson axial vertex to neutrons dnA")

    parser.add_argument("--dprotonS", type=float, help="SM Z boson vector vertex to protons dpS")
    parser.add_argument("--dneutronS", type=float, help="SM Z boson axial vertex to neutrons dnS")

    parser.add_argument("--dprotonP", type=float, help="SM Z boson vector vertex to protons dpP")
    parser.add_argument("--dneutronP", type=float, help="SM Z boson axial vertex to neutrons dnP")

    # parser.add_argument("--dprotonV", type=float, help="Z' vector vertex to charged leptons deV")
    # parser.add_argument("--dneutronA", type=float, help="Z' axial vertex to charged leptons deA")

    # parser.add_argument("--dprotonS", type=float, help="h' scalar vertex to charged leptons dSe")
    # parser.add_argument("--dneutronP", type=float, help="h' pseudoscalar vertex to charged leptons dPe")


def add_scope_arguments(parser, DEFAULTS):

    ##### visible final states in HNL decay
    parser.add_argument("--decay_product", type=str.lower, help="decay process of interest", choices=DEFAULTS._choices["decay_product"])

    ##### experiments
    parser.add_argument("--exp", type=str.lower, help="experiment file path or keyword")

    ##### scattering types
    parser.add_argument("--nopelastic", help="do not generate proton elastic events", action="store_true")
    parser.add_argument("--nocoh", help="do not generate coherent events", action="store_true")
    parser.add_argument("--include_nelastic", help="generate neutron elastic events", action="store_true")

    parser.add_argument("--noHC", help="do not include helicity conserving events", action="store_true")
    parser.add_argument("--noHF", help="do not include helicity flipping events", action="store_true")

    # projectile neutrino flavors
    parser.add_argument("--nu_flavors", type=str.lower, action="store", nargs="+", choices=DEFAULTS._choices["nu_flavors"])

    # enforce the parent HNL decay to be prompt
    parser.add_argument("--enforce_prompt", help="forces the particles to decay promptly", action="store_true")
    parser.add_argument("--make_summary_plots", help="generate summary plots of kinematics", action="store_true")


def add_mc_arguments(parser, DEFAULTS):

    ##### run related arguments
    parser.add_argument("--loglevel", type=str.lower, help="Logging level")
    parser.add_argument("--verbose", help="Verbose for logging", action="store_true")
    parser.add_argument("--logfile", type=str, help="Path to logfile. If not set, use std output.")

    ##### Vegas parameters
    parser.add_argument("--neval", type=int, help="number of evaluations of integrand")
    parser.add_argument("--nint", type=int, help="number of adaptive iterations")
    parser.add_argument("--neval_warmup", type=int, help="number of evaluations of integrand in warmup")
    parser.add_argument("--nint_warmup", type=int, help="number of adaptive iterations in warmup")

    ##### program options
    parser.add_argument("--pandas", help="If true, prints pandas dataframe to .pckl files", action="store_true")
    parser.add_argument("--parquet", help="If true, prints pandas dataframe to .parquet files. Loses metadata in attrs.", action="store_true")
    parser.add_argument("--numpy", help="If true, prints events as ndarrays in a .npy file", action="store_true")
    parser.add_argument("--hepevt", help="If true, print events to HEPEVT-formatted text files (does not save event weights)", action="store_true")
    parser.add_argument("--hepevt_legacy",
        help="If true, print events to a legacy HEPEVT format (saving weights next to the number of particle in the event and without linebreaks in particle entries)",
        action="store_true",
    )
    parser.add_argument("--hepmc2", help="If true, prints events to HepMC2 format.", action="store_true")
    parser.add_argument("--hepmc3", help="If true, prints events to HepMC3 format.", action="store_true")
    parser.add_argument("--hep_unweight", help="unweight events when printing in standard HEP formats (needs large neval)", action="store_true")
    parser.add_argument("--unweighted_hep_events",type=int,
        help="number of unweighted events to accept in any of the standard HEP formats. Has to be much smaller than neval for unweight procedure to work.",
    )

    parser.add_argument("--sparse", help="drop all information in the event except for visible particle momenta, neutrino energy, and weights.", action="store_true"
    )
    parser.add_argument("--print_to_float32", help="Use float32 instead of default float64 when printing output to save storage space.", action="store_true")

    parser.add_argument("--path", type=str, help="path where to save run's outputs")
    parser.add_argument("--seed", type=int, help="numpy seed to be used by vegas.")
